#!/usr/bin/env python

__author__ = 'jsommers@colgate.edu'
__doc__ = '''
Capture daemon proxy server to handle XML/RPC calls from GIMS UI-side process. 
Handles managing experiment state, starting, configuring, and stopping
capture processes, and managing data upload.

This source code is licensed under the GENI public license.
See www.geni.net, or "geni_public_license.txt" that should 
have accompanied this software.
'''


import sys
from SimpleXMLRPCServer import SimpleXMLRPCServer
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler
import xmlrpclib
from optparse import OptionParser
import multiprocessing
import Queue
import re
import os
import datetime
import signal
import time
import logging
import random
import StringIO
import capd_storage


# the lone global variable; flag to tell whether we should quit the
# program or not.
running = True


class RequestHandler(SimpleXMLRPCRequestHandler):
    '''
    Overridden base xmlrpc request handler.  Just modify the logging
    to flow into our own logger.
    '''
    rpc_paths = ('/RPC2', '/')

    def log_message(self, format, *args):
        pre = 'xmlrpc request from ' + self.address_string() + ' '
        xstr = format % (args)
        logger = logging.getLogger('capd_proxy')
        logger.debug(pre + xstr)


class ExperimentState(object):
    '''
    Container class to encapsulate information about an individual
    experiment.
    '''
    def __init__(self, exptname, xdir):
        self.experiment_name = exptname
        self.capture_on = False
        self.capture_pid = -1
        self.config = dict()
        self.exptdir = xdir
        self.timeidle = time.time()
        self.proc = None

    def getId(self):
        return self.experiment_name

    def getAge(self):
        if self.capture_on:
            return 0
        return int(time.time() - self.timeidle)

    def setConfig(self, key, value):
        self.config[key] = value

    def getConfig(self, key):
        return self.config.get(key, None)

    def clean(self):
        if self.capture_on and self.capture_pid > 0:
            os.kill(pid, signal.SIGINT)
 
    def setActive(self, b):
        self.timeidle = time.time()
        self.capture_on = b

    def getActive(self):
        return self.capture_on

    def __str__(self):
        rv = 'Experiment: ' + self.experiment_name + ' :-: '
        kvlist = []
        for k,v in self.config.items():
            kvlist.append(k + '->' + str(v))
        rv += ':::'.join(kvlist)
        rv += ' :-: capture on ' + str(self.capture_on)
        rv += ' :-: capture pid ' + str(self.capture_pid)
        return rv

    def getStorageInfo(self):
        stype = 'local'
        stup = (None)
        
        if 'sshuser' in self.config and 'sshhost' in self.config and 'sshkey' in self.config:
            stype = 'ssh'
            slist = []
            slist.append(self.config['sshuser'])
            slist.append(self.config['sshhost'])
            slist.append(self.config['sshkey'])
            if 'sshpath' in self.config:
                slist.append(self.config['sshpath'])
            if 'sshport' in self.config:
                slist.append(self.config['sshport'])
            stup = tuple(slist)
            
        elif 's3key' in self.config and 's3secret' in self.config and 's3bucket' in self.config:
            stype = 's3'
            stup = (self.config['s3key'], self.config['s3secret'], self.config['s3bucket'])

        return stype, self.exptdir, stup
        

class ExperimentManager(object):
    '''
    The main class of this program: all the xmlrpc methods are defined
    here, and all the important state is stored here.
    '''

    def __init__(self, logger, debug, basedir, defaultdev, capdexec, tmo=3600):
        self.exptdb = dict()
        self.piddb = dict()
        self.logger = logger
        self.debug = debug
        self.basedir = basedir
        self.defaultdev = defaultdev
        self.capdexec = capdexec
        self.testfile = 'GIMStestfile.txt'
        self.expttimeout = tmo

        # create file for testing uploads
        tout = open(self.testfile, 'w')
        print >>tout, "This is a dummy file for testing uploads."
        tout.close()

        # install special signal handlers after storage
        # processes have been started
        signal.signal(signal.SIGINT, self.sigInt)
        signal.signal(signal.SIGTERM, self.sigInt)
        signal.signal(signal.SIGPIPE, signal.SIG_IGN)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        self.storage_daemon = None
        self.startStorageAgent()


    def startStorageAgent(self):
        toc,fromc = multiprocessing.Queue(),multiprocessing.Queue()
        p = multiprocessing.Process(target=capd_storage.storage_agent, args=(toc, fromc, self.basedir, self.debug))
        self.storage_daemon = (p, toc, fromc)
        p.start()

        
    def stopStorageAgents(self):
        self.logger.info("Stopping storage agent.")
        self.storage_daemon[1].put(('STOP',))
        try:
            self.storage_daemon[0].terminate()
            self.storage_daemon[0].join(1)
        except OSError,e:
            pass
        self.logger.debug("After stopping storage agent.")


    def sigInt(self, signum, stackframe):
        global running
        self.logger.info('Caught sig %d in RPC listener.  Going down.' % (signum))
        running = False
    

    def maintenance(self):
        deadexpts = []
        for eid,exptstate in self.exptdb.items():
            if exptstate.getAge() > self.expttimeout:
                self.logger.info('Cleaning up stale experiment state for ' + eid + ' ' + str(exptstate))
                exptstate.clean()
                deadexpts.append(eid)
        for eid in deadexpts:
            del self.exptdb[eid]


    def getState(self):
        outstr = 'Experiments:\n'
        for k,v in self.exptdb.items():
            outstr += str(k) + ' ' + str(v) + '\n'
        return outstr


    def TestExperimentStorage(self, params):
        '''
        Alias for testExperimentStorage
        '''
        return self.testExperimentStorage(params)


    def testExperimentStorage(self, params):
        '''
        Test whether experiment storage settings result in
        correct uploading.
        params: exptid (string)
        return: response string, or xmlrpc.fault on error
        '''
        self.maintenance()
        self.checkParams('startExperiment', params)

        exptid = self.getExptId(params)

        if exptid not in self.exptdb:
            self.logger.info('Throwing exception: no such experiment: '+exptid)
            raise xmlrpclib.Fault(-1, 'No such experiment '+exptid+' exists')
        self.logger.info('Attempt to test storage for experiment %s' % (exptid))

        expt = self.exptdb[exptid]
        self.checkSufficientConfig(expt) 

        stype,xdir,stup = expt.getStorageInfo()
        self.logger.info('Testing storage %s for experiment %s cred(%s)' % (stype, expt.getId(), str(stup)))


        flist = [(self.testfile, self.testfile)]
        try:
            response = 'Local storage ok.'
            if stype == 's3':
                capd_storage.s3_upload(self.logger, exptid, stup[0], stup[1], stup[2], flist)
                response = 's3 storage appears to work.  please verify upload of GIMStest.txt'
            elif stype == 'ssh':
                xport = 22
                xpath = ''
                if len(stup) > 3:
                    try:
                        xport = int(stup[3])
                    except:
                        xpath = stup[3]

                    if len(stup) > 4:
                        xport = int(stup[4])

                newflist = [ ('GIMStest.txt', xpath + 'GIMStest.txt') ]
                capd_storage.sftp_upload(self.logger, exptid, stup[1], stup[0], stup[2], newflist, xport)
                response = 'ssh storage appears to work.  please verify upload of GIMStest.txt'
            elif stype == 'local':
                local_upload(self.logger, exptid, flist)
        except Exception,e:
            raise xmlrpclib.Fault(-1, 'Error while testing storage: '+str(e))

        self.logger.info('Storage test result for %s: %s' % (exptid, response))
        return response
      


    def StartExperiment(self, params):
        '''
        Alias for startExperiment.
        '''
        return self.startExperiment(params)


    def startExperiment(self, params):
        '''
        params: exptid (string)
        return: response string or xmlrpc.Fault on error
        '''
        self.maintenance()
        self.checkParams('startExperiment', params)

        exptid = self.getExptId(params)

        if exptid not in self.exptdb:
            self.logger.info('Throwing exception: no such experiment: '+exptid)
            raise xmlrpclib.Fault(-1, 'No such experiment '+exptid+' exists')
        self.logger.info('Attempt to start experiment %s' % (exptid))

        expt = self.exptdb[exptid]
        self.checkSufficientConfig(expt)
        pid = self.doStartExperiment(expt)
        return ''.join(['Successfully started experiment ',exptid,'.  Process id ',str(pid)])


    def StopExperiment(self, params):
        '''
        Alias for stopExperiment.
        '''
        return self.stopExperiment(params)


    def stopExperiment(self, params):
        '''
        params: exptid(string)
        return: response string or xmlrpc.Fault
        '''
        self.maintenance()
        self.checkParams('stopExperiment', params)

        exptid = self.getExptId(params)
        if exptid not in self.exptdb:
            self.logger.info('Throwing exception: no such experiment: '+exptid)
            raise xmlrpclib.Fault(-1, 'No such experiment '+exptid+' exists')

        self.logger.info('Attempt to stop experiment %s' % (exptid))

        expt = self.exptdb[exptid]
        pid = self.doStopExperiment(expt)
        return ''.join(['Successfully stopped experiment ',exptid,', process id',str(pid)])


    def GetExperimentStatus(self, params):
        '''
        Alias for getExperimentStatus.
        '''
        return self.getExperimentStatus(params)


    def getExperimentStatus(self, params):
        '''
        params: exptid (string)
        return: all expt params + CaptureStatus (as a string) or
        xmlrpc.Fault
        '''
        self.maintenance()
        self.checkParams('getExperimentStatus', params)

        exptid = self.getExptId(params)
        if exptid not in self.exptdb:
            self.logger.info('Throwing exception: no such experiment: '+exptid)
            raise xmlrpclib.Fault(-1, 'No such experiment '+exptid+' exists')

        self.logger.info('Attempt to obtain experiment info %s' % (exptid))

        expt = self.exptdb[exptid]
        st = self.gatherConfig(expt)

        return str(st)


    def ConfigureExperiment(self, params):
        '''
        Alias for configureExperiment.
        '''
        return self.configureExperiment(params)


    def configureExperiment(self, params):
        '''
        params: exptid (string), --many other params-- (see doConfig)
        return: status (string) or xmlrpc.Fault on error
        '''
        self.maintenance()
        self.checkParams('configureExperiment', params)

        exptid = self.getExptId(params)
        self.logger.debug('Configure experiment %s' % (exptid))
        if exptid not in self.exptdb:
            estate = ExperimentState(exptid, self.basedir + os.path.sep + exptid)
            self.exptdb[exptid] = estate
            self.logger.info(' '.join(['New experiment state created for',str(exptid),str(params)]))

        expt = self.exptdb[exptid]
        warnings = self.doConfig(expt, params)
        wstr = 'Success'
        if len(warnings):
            wstr = ':::'.join(warnings)
        self.logger.info('Configure experiment %s.  Current state: %s; config warnings: %s' % (exptid, str(expt), wstr))
        return ''.join(['Successfully configured experiment ',exptid,'. Current state: ',str(expt),'.  Configuration warnings: ',wstr])

    
    def checkParams(self, methodname, params):
        if not self.getExptId(params):
            self.logger.info('Throwing exception: no experiment specified')
            raise xmlrpclib.Fault(-1, 'Missing \'ExperimentId\' in parameter structure')
        if methodname == 'startExperiment':
            return None 
        elif methodname == 'stopExperiment':
            return None 
        elif methodname == 'configureExperiment':
            return None 
        elif methodname == 'getExperimentStatus':
            return None 
        else:
            self.logger.info('Throwing exception: invalid method name')
            raise xmlrpclib.Fault(-1, 'Invalid method name in checkParams')
    

    def checkSufficientConfig(self, exptstate):
        if not exptstate.getId():
            self.logger.info('Throwing exception: missing experiment id')
            raise xmlrpclib.Fault(-1, 'Missing Experiment Id')
        stype = exptstate.getConfig('storage')
        if not stype:
            self.logger.info('Throwing exception: no storage configuration')
            raise xmlrpclib.Fault(-1, 'No storage configuration')
        if stype == 's3':
            if not exptstate.getConfig('s3key'):
                raise xmlrpclib.Fault(-1, 'Missing s3 access key')
            elif not exptstate.getConfig('s3secret'):
                raise xmlrpclib.Fault(-1, 'Missing s3 secret key')
            elif not exptstate.getConfig('s3bucket'):
                raise xmlrpclib.Fault(-1, 'Missing s3 bucket name')
        elif stype == 'ssh':
            if not exptstate.getConfig('sshuser'):
                raise xmlrpclib.Fault(-1, 'Missing SSH user')
            elif not exptstate.getConfig('sshhost'):
                raise xmlrpclib.Fault(-1, 'Missing SSH hostname')
            elif not exptstate.getConfig('sshkey'):
                raise xmlrpclib.Fault(-1, 'Missing SSH private key')
        stype = exptstate.getConfig('sampletype')
        if stype and stype != 'none':
            sparm = exptstate.getConfig('sampleparam')
            if not sparm:
                raise xmlrpclib.Fault(-1, 'Missing sample parameter')
        atype = exptstate.getConfig('anontype')
        if atype and atype != 'none':
            aparm = exptstate.getConfig('anonkey')
            if atype != 'anonymize' or not aparm:
                # FIXME: need to really raise the error; need a key!
                exptstate.setConfig('anonkey', 'abcdabcdabcdabcd')
                # raise xmlrpclib.Fault(-1, 'Incomplete anonymization configuration')
        return None
    
    
    def constructCommandLine(self, exptstate):
        cmdlist = [ self.capdexec ]
        if exptstate.getConfig('device'):
            if 'eth' in exptstate.getConfig('device'):
                cmdlist.append("-d")
                cmdlist.append(exptstate.getConfig('device'))
            else:
                cmdlist.append('-d')
                cmdlist.append(self.defaultdev)
        elif len(self.defaultdev):
            cmdlist.append('-d')
            cmdlist.append(self.defaultdev)

        filterexpr = ''
        if exptstate.getConfig('vlan'):
            xvlan = -1
            try:
                xvlan = int(exptstate.getConfig('vlan'))
            except:
                pass
            if xvlan > 0: 
                filterexpr += 'vlan ' + str(exptstate.getConfig('vlan'))
        if exptstate.getConfig('filterexpr'):
            if len(filterexpr):
                filterexpr += ' and'
            filterexpr += ' ' + exptstate.getConfig('filterexpr')
        if len(filterexpr):
            cmdlist.append('-s')
            cmdlist.append(filterexpr)

        outdir = self.basedir + os.path.sep + exptstate.getId()
        self.logger.debug('Making output directory: '+outdir)
        try:
            os.makedirs(outdir)
        except OSError,e:
            if 'File exists' in str(e):
                pass
            else:
                self.logger.error('Error making temp storage pen: ' + str(e))
                raise xmlrpclib.Fault(-1, 'Error making local storage for capture files:' + str(e))
                
        cmdlist.append('-p')
        cmdlist.append(outdir)

        siteloc = exptstate.getConfig('sitelocation')
        if siteloc:
            cmdlist.append('-l')
            cmdlist.append(siteloc)

        metatext = exptstate.getConfig('metatext')
        if metatext:
            cmdlist.append('-u')
            cmdlist.append(metatext)

        cmdlist.append('-N')
        cmdlist.append(exptstate.getId())

        # sampling
        samptype = exptstate.getConfig('sampletype')
        if samptype and samptype != 'none':
            cmdlist.append('-t')
            if samptype == 'everyN':
                cmdlist.append('3')
            elif samptype == 'uniformrandom':
                cmdlist.append('2')
            cmdlist.append('-r')
            sampparam = exptstate.getConfig('sampleparam')
            cmdlist.append(sampparam)

        # anonymization
        anontype = exptstate.getConfig('anontype')
        if anontype and anontype != 'none':
            cmdlist.append('-k')
            anonkey = exptstate.getConfig('anonkey')
            cmdlist.append(anonkey)

        # aggregation
        aggtype = exptstate.getConfig('aggregationtype')
        if aggtype and aggtype != 'none':
            cmdlist.append('-a')
            if aggtype == 'count_pkts':
                cmdlist.append('count')
            elif aggtype == 'combine_pkt_flows':
                cmdlist.append('flow')

        # file rollover period
        cmdlist.append('-z')
        rolloverint = exptstate.getConfig('rollover')
        if rolloverint:
            cmdlist.append(str(int(rolloverint)*60))
        else:
            cmdlist.append('30') # default to 30 seconds

        # ensure that everything is a string that we pass
        # to the cmd line.
        cmdlist = [ str(x) for x in cmdlist ]

        return self.capdexec, outdir, cmdlist
    
    
    def notifyStorageStart(self, exptstate):
        stype,xdir,stup = exptstate.getStorageInfo()
        q = self.storage_daemon[1]
        self.logger.info('Notifying storage agent about experiment %s (%s)' % (exptstate.getId(), stype))
        q.put( ('NEWEXPT', exptstate.getId(), stype, stup, xdir) )


    def notifyStorageStop(self, exptstate):
        stype,xdir,stup = exptstate.getStorageInfo()
        q = self.storage_daemon[1]
        self.logger.info('Notifying storage to finish experiment %s (%s)' % (exptstate.getId(), stype))
        q.put( ('RMEXPT', exptstate.getId(), stype, stup, xdir) )


    def reapChildren(self):
        deadpids = []
        for pid in self.piddb:
            exptstate = self.piddb[pid]
            if exptstate.proc and not exptstate.proc.is_alive():
                self.logger.debug('Reaping child process %d' % (pid))
                deadpids.append(pid)
                exptstate.setActive(False)
                exptstate.proc.join()
                exptstate.proc = None
        for pid in deadpids:
            del self.piddb[pid]


    def _capdentry(self, path, args):
        '''
        Entrypoint for capture daemon binary startup
        '''
        try:
            os.execv(path, args)
        except OSError,e:
            self.logger.warn('Child process failed to start: %s' % (str(e)))


    def doStartExperiment(self, exptstate):
        '''
        Start an experiment.  Assumes that configuration has been checked
        for sufficiency.
        '''
        path,outdir,args = self.constructCommandLine(exptstate)

        self.notifyStorageStart(exptstate)
        self.logger.info('Starting experiment: %s // %s // %s' % (path,outdir,str(args)))

        p = multiprocessing.Process(target=self._capdentry, args=(path, args))
        p.start()
        # FIXME: p may not be *immediately* alive.  need a better way to ensure
        # experiment has started up
        if not p.is_alive():
            self.logger.info('Throwing exception: cannot spawn capture process: ' + str(e))
            raise xmlrpclib.Fault(-1, 'Couldn\'t spawn capture process for experiment %s: %s' % (exptstate.getId(), str(e)))
        exptstate.proc = p
        exptstate.setActive(True)
        self.piddb[p.pid] = exptstate
        return p.pid

    
    def doStopExperiment(self, exptstate):
        pid = exptstate.proc.pid
        if pid > 0:
            self.logger.info('Sending SIGTERM to capture process %d.' % (pid))
            exptstate.proc.terminate()
            exptstate.setActive(False)
            return
        else:
            exptstate.setActive(False)
            raise xmlrpclib.Fault(-1, 'Experiment is not running')
        self.notifyStorageStop(exptstate)
        return pid

    
    def gatherConfig(self, exptstate):
        return str(exptstate)

    
    def getExptId(self, params):
        rv = None
        if 'ExperimentId' in params:
            rv = params['ExperimentId'] 
        elif 'ExperimentID' in params:
            rv = params.pop('ExperimentID')
            params['ExperimentId'] = rv
        return rv
    
    
    def doStorageConfig(self, exptstate, xdict):
        '''
        StorageSpec (struct)
           storagetype (string) ("local", "s3", "ssh")
           storageparams (struct)
                local: local_storage_dir
                s3: s3accesskey: string, s3secretkey: string, s3bucket: string
                ssh: sshuser: string, sshhost: string, sshport: int, sshpath: string, sshkey: string
     
                  // default: either local or no default (i.e., this is required)
                  // "local" is probably something we only want certain users to
                  // be able to do, since the user would have to know how/where to
                  // retrieve the data.
        '''
        warnings = []
        stype = ''
    
        for skey in xdict.keys():
            if skey.lower() == 'storagetype':
                stype = xdict[skey]
                if stype.lower() not in ['local', 's3', 'ssh']:
                    warnings.append('Storage type ' + str(stype) + ' unrecognized.')
                else:
                    stype = exptstate.setConfig('storage', stype.lower())
            elif skey.lower() == 'storageparams':
                sparam = xdict['storageparams']
                for k in sparam.keys():
                    if k == 's3key':
                        exptstate.setConfig('s3key', sparam[k])
                    elif k == 's3bucket':
                        exptstate.setConfig('s3bucket', sparam[k])
                    elif k == 's3secret':
                        exptstate.setConfig('s3secret', sparam[k])
                    elif k == 'sshuser':
                        exptstate.setConfig('sshuser', sparam[k])
                    elif k == 'sshhost':
                        exptstate.setConfig('sshhost', sparam[k])
                    elif k == 'sshport':
                        exptstate.setConfig('sshport', sparam[k])
                    elif k == 'sshpath':
                        exptstate.setConfig('sshpath', sparam[k])
                    elif k == 'sshkey':
                        exptstate.setConfig('sshkey', sparam[k])
                    elif k == 'local_storage_dir':
                        exptstate.setConfig('localdir', sparam[k])
                    elif k == 'rollover_interval':
                        rollint = sparam[k]
                        try:
                            rollint = int(rollint)
                        except:
                            rollint = 1
                        exptstate.setConfig('rollover', rollint)
                    else:
                        warnings.append('Unrecognized storage param: ' + str(k))
        return warnings
    
    
    def doMetaDataConfig(self, exptstate, xdict):
        '''
        MetaDataSpec (struct)
           UserText (string)     // user-defined description to attach with the
                                 // data so that he/she has their own note as to
                                 // what went into the experiment
                                 // default: no user metadata
     
        '''
        warnings = []
        for mdkey in xdict.keys():
            if mdkey.lower() == 'usertext':
                exptstate.setConfig('metatext', xdict[mdkey])
            else:
                warnings.append('Unrecognized metadata param ' + str(mdkey))
        return warnings
    
    
    def doCaptureConfig(self, exptstate, xdict):
        '''
        CaptureSpec (struct)
           filterexpr (string)   // e.g., pcap expression
           device (string)       // specific physical device name
                                 // default: default device and no filter expression
        '''
        warnings = []
        for key in xdict.keys():
            self.logger.debug('captureconfig key %s -> %s' % (key, xdict[key]))
            if key.lower() == 'filterexpr':
                exptstate.setConfig('filterexpr', xdict[key])
            elif key.lower()  == 'device':
                exptstate.setConfig('device', xdict[key])
            else:
                warnings.append('Unrecognized capture param ' + str(key))
        return warnings
    
    
    def doTransformConfig(self, exptstate, xdict):
        '''
        TransformSpec (struct)   
           SampleSpec (struct)
               sampletype (string) "everyN", or "uniformrandom"
               sampleparams (array of strings?)
                  // types: "everyN", "uniformrandom", 
                  // for these two sample types, there's only one parameter
                  // necessary, but other sample types (e.g., trajectory sampling)
                  // might have more than one parameter necessary.  to pass those
                  // parameters we might either have an array of strings, or perhaps
                  // better, we could have named elements in the SampleSpec structure
                  // for each necessary parameter.  e.g., "N" (if "everyN" is the type),
                  // or "p" if the type is "uniformrandom"
     
                  // default: no transformation
     
           AnonSpec (struct)
               anontype (string) 'anonymize'
               anonkey (string) 
     
           AggregationSpec (struct)
               aggtype (string)  'combine_pkts_flows', 'combine_pkts'
               aggparams (array of strings?)
                  // no need to define these yet, but for aggregation into 
                  // flow records, we might have parameters to define specifically
                  // what constitutes a flow (e.g., 5-tuple or some other combo)
     
                  // default: no aggregation
        '''
        warnings = []
        for key in xdict.keys():
            self.logger.debug('Transform spec key %s -> %s' % (key, str(xdict[key])))
            if key.lower() == 'samplespec':
                sampdict = xdict[key]
                for skey in sampdict.keys():
                    self.logger.debug('samplespec key %s -> %s' % (skey, str(sampdict[skey])))
                    if skey.lower() == 'sampletype':
                        stype = sampdict[skey]
                        if stype.lower() in ['everyn', 'uniformrandom', 'none']:
                            exptstate.setConfig('sampletype', sampdict[skey])
                        else:
                            warnings.append('Unrecognized sample type ' + str(skey))
                    elif skey.lower() == 'sampleparams':
                        sparams = sampdict[skey]
                        for spkey in sparams.keys():
                            if spkey.lower() == 'n' or spkey.lower() == 'p':
                                sparm = -1.0
                                try:
                                    sparm = float(sparams[spkey])
                                except:
                                    pass
                                if sparm < 0:
                                    warnings.append('Invalid sample parameter ' + str(sparams[spkey]))
                                else:
                                    exptstate.setConfig('sampleparam', sparm)
                    else:
                        warnings.append('Unrecognized sample parameter ' + str(skey))
                    
            elif key.lower() == 'anonspec':
                anondict = xdict[key]
                for akey in anondict.keys():
                    if akey.lower() == 'anontype':
                        exptstate.setConfig('anontype', anondict[akey])
                    elif akey.lower() == 'anonkey':
                        exptstate.setConfig('anonkey',anondict[akey])
                    else:
                        warnings.append('Unrecognized anonymization parameter ' + str(akey))
    
            elif key.lower() == 'aggregationspec':
                aggdict = xdict[key]
                for akey in aggdict.keys():
                    if akey.lower() == 'aggtype':
                        exptstate.setConfig('aggregationtype', aggdict[akey])
                    else:
                        warnings.append('Unrecognized aggregation parameter ' + str(akey))
            else:
                warnings.append('Unrecognized transformation type ' + str(key))
    
        return warnings
    

    def doConfig(self, exptstate, cdict):
        '''
        input: dictionary with configuration sections:
        - VLAN
        - MetaDataSpec
        - CaptureSpec
        - TransformSpec
        - StorageSpec
        (may want a CompressionSpec eventually)
        return list of warning strings
        '''
        warnings = []
    
        for key in cdict.keys():
            self.logger.debug('Configure key %s -> %s' % (key, str(cdict[key])))
            
            if key.lower() == 'vlan':
                vlan = -1
                try:
                    vlan = int(cdict[key])
                except:
                    pass
                exptstate.setConfig('vlan', vlan)
            elif key.lower() == 'metadataspec':
                warnings += self.doMetaDataConfig(exptstate, cdict[key])
            elif key.lower() == 'capturespec':
                warnings += self.doCaptureConfig(exptstate, cdict[key])
            elif key.lower() == 'transformspec':
                warnings += self.doTransformConfig(exptstate, cdict[key])
            elif key.lower() == 'anonspec':
                warnings += self.doTransformConfig(exptstate, {'AnonSpec':cdict[key]})
            elif key.lower() == 'aggregationspec':
                warnings += self.doTransformConfig(exptstate, {'AggregationSpec':cdict[key]})
            elif key.lower() == 'storagespec':
                warnings += self.doStorageConfig(exptstate, cdict[key])
            elif key.lower() == 'experimentid':
                pass
            elif key.lower() == 'methodname':
                pass
            elif key.lower() == 'sitelocation':
                exptstate.setConfig('sitelocation', cdict[key].upper())
            elif key.lower() == 'devicename':
                exptstate.setConfig('devicehost', cdict[key])
            else:
                warnings.append('Unrecognized configuration section ' + str(key))
        self.logger.debug('Exiting configuration routine.')
        return warnings


def main():
    parser = OptionParser()
    parser.prog = "capd_proxy"
    parser.add_option("-p", "--port", dest="port", default=8001,
                      type="int", action="store", help="Set XML/RPC listen port")
    parser.add_option("-l", "--host", dest="host", default="0.0.0.0",
                      action="store", help="Set host/IP to which XML/RPC listener is bound")
    parser.add_option("-x", "--debug", dest="debug", default=False,
                      action="store_true", help="Turn on debugging output")
    parser.add_option("-b", "--basedir", dest="basedir", default="./capdata",
                      help="Set the base directory for output measurements")
    parser.add_option("-d", "--device", dest="defaultdevice", default="eth0",
                      help="Set the default capture device")
    parser.add_option("-c", "--capdaemon", dest="capdexec", 
                      # default="./fake-capture-daemon",
                      default="./capture-daemon/capture-daemon",
                      help="Set the path to the capture-daemon executable")
    (options, args) = parser.parse_args()



    if options.debug:
        logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(name)-14s %(levelname)-8s %(message)s', filename='capd_proxy.log')
    else:
        logging.basicConfig(level=logging.INFO, format='%(asctime)s %(name)-14s %(levelname)-8s %(message)s', filename='capd_proxy.log')

    console = logging.StreamHandler() 
    format = logging.Formatter('%(asctime)s %(name)-14s %(levelname)-8s %(message)s')
    console.setFormatter(format)
    logging.getLogger('').addHandler(console)
    
    logger = logging.getLogger('capd_proxy')
    logger.info('capd_proxy starting up.')

    # actives = ActiveExperiments()
    # handler = ExperimentManager(actives, logger, options.debug, options.basedir, 
    handler = ExperimentManager(logger, options.debug, options.basedir, 
                                options.defaultdevice, options.capdexec)

    server = SimpleXMLRPCServer((options.host, options.port), requestHandler=RequestHandler)
    server.register_introspection_functions()

    server.register_function(handler.startExperiment)
    server.register_function(handler.StartExperiment)
    server.register_function(handler.stopExperiment)
    server.register_function(handler.StopExperiment)
    server.register_function(handler.configureExperiment)
    server.register_function(handler.ConfigureExperiment)
    server.register_function(handler.getExperimentStatus)
    server.register_function(handler.GetExperimentStatus)
    server.register_function(handler.testExperimentStorage)
    server.register_function(handler.TestExperimentStorage)
    server.register_function(handler.getState)

    server.timeout = 1.0

    global running
    while running:
        try:
            # server.serve_forever()
            server.handle_request()
            handler.reapChildren()
        except Exception,e:
            logger.debug('Caught exception while serving XML/RPC requests: '+str(e)) 
    logger.debug("proxy out of main loop")

    handler.stopStorageAgents()

    nchildren = 0
    try:
        nchildren = len(multiprocessing.active_children())
    except OSError,e:
        if 'No child process' in e:
            nchildren = 0

    logger.debug('Active children remaining: %d' % (nchildren))
    while nchildren > 0:
        try:
            for p in multiprocessing.active_children():
                p.terminate()
                p.join(1)
        except OSError,e:
            if 'No child process' in e:
                nchildren = 0
    logger.info('No more child processes alive --- going down.')

if __name__ == '__main__':
    main()
